/*
	wait_gpio.c - gpio-тест ожидания прерываний на заданных линиях
    разработчик - Т.Гумер (gumer@mcst.ru).

    автономная сборка :
gcc -g -o wait_gpio wait_gpio.c
или с контролем...
gcc -g -Wall -o wait_gpio wait_gpio.c
gcc -g -Wunused-variable -o wait_gpio wait_gpio.c

структура файла = <парам-файл> :
    который определяет пины исключаемые из тестирования
    и описывает пары пинов проверки соединений.

NGPIO	16
-e '-4, 8-15, 17, 29-31'
    #	то же
-e "0-4, 8-15, 17, 29-31"
EOF

   что бы увидеть (при отладке) получения событий от 2-х прерываний :
./wait_gpio -p CMDone -P 8 -p CMDone -P 10 -e '0-6 11-' -t 30 -www -s -S 130

   запуск с подыгрывающим скриптом CMD, который выдает последовательности сигналов :
	    SET 7 0; sleep 1.3; SET 7 1; sleep 1.3
	    SET 9 0; sleep 1.3; SET 9 1; sleep 1.3
./wait_gpio -p CMD -P 7 -P 9 -e '0-6 11-' -t 30 -www -s
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <poll.h>
#include <ctype.h>

#define NM_TST  "Тест ожидания прерываний модуля GPIO"
//	описания из gpio.h

#define MAX_NUM_GPIOS   256     // мах число GPIO на ВК
#define MAX_NUM_CHECKS  512     // мах число атомарных проверок

#define MEM_ERR         (-1)
#define FILE_ERR        (-2)
#define SIGN_ERR        (-3)
#define INPUT_ERR       (-4)

#define MAX_SYSFS_NAME_LEN      120
#define MAX_COMMAND_LEN         10
#define DIRECTION_IN            0
#define DIRECTION_OUT           1
#define LEVEL_LOW               0
#define LEVEL_HIGH              1

/*
 * Элемент описывающий линию GPIO (пин).
 */
typedef struct circuit {
        int never_export;       // исключен из проверок согласно параметров
        int never;              // исключен из проверок
        int avail;              // пин exported/unexported
        int active_low;         // "0" - signal is "0", "1" - signal is "1"
        int solder;             // пин запаян
        int no_open;            // не открывается
        int direction;          // =-1 - не определено, =0 - IN, =1 - OUT
} circuit_t;

//end	~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ описания из gpio.h

static char pins_name[16] = "gpio";

static int NGPIO = -1;		// число ngpio из `параметров`
static int ngpio = -1;		// число ngpio из `sysfs-интерфейса`
static int NPOLFD = 0;		// число массива pollfd для операций poll
static int wait_time = 0;	// > 0 - время ожидания прерываний в секундах

static char my_dir_name[PATH_MAX];
static char *my_prog_name = NULL;

#define DGPIO "/sys/class/gpio"
#define USE_GETTIMEOFDAY	1
#define HRT_SEC(SEC)	((hrtime_t)(SEC *1000000000LL))
#define TST_FAIL	-1
#define TST_PASS	0
#define NO_CNCT		-9999

static circuit_t	pins[MAX_NUM_GPIOS];	// таблица линий связи

static int	verbose = 0;
#define	vverbose	(verbose > 1)
#define	vvverbose	(verbose > 2)

		//	по умолчанию будет =1 - т.е. по переднему фронту
static int     wait_regime = 0;        // > 0 - режим ожидания прерываний

#define	FNC	__FUNCTION__
#define	LN	__LINE__

static int	my_pid;
static int	errors = 0;		// суммарное число ошибок
static int	errpoll = 0;		// сколько poll с revents ? POLLERR
static int	limit_error = 20;	// число ошибок по-уммолчанию
static int	limit_errpoll = 2;	// число ошибок errpoll по-уммолчанию
static int	without_check_solder = 0; // без контроля запаяных пинов
static int	timeout_before_check = 0; // задержка в милисек. перед проверкой
static int	*pin2poll = NULL;	// таблица для перевода N пина в N pollfd
static int	*poll2pin = NULL;	// таблица для перевода N pollfd в N пина
static int	*pin2fd = NULL;		// таблица для хранения fd пин-файлов
static struct	pollfd *pollfd = NULL;	// pollfd для операций poll

/*
    #define POLLIN      0x0001    Можно считывать данные
    #define POLLPRI     0x0002    Есть срочные данные
    #define POLLOUT     0x0004    Запись не будет блокирована
    #define POLLERR     0x0008    Произошла ошибка
    #define POLLHUP     0x0010   "Положили трубку"
    #define POLLNVAL    0x0020    Неверный запрос: fd не открыт
 */
static int	POLL_CH	= POLLPRI;	// POLLPRI | POLLIN | POLLOUT
//int	POLL_CH	= POLLIN;	// POLLPRI | POLLIN | POLLOUT - не работает !!!

/*
 * число циклов проверки соединений, = -1 проверки по времени
 */

static int	number_checks = -1; // для отладки - предел числа элементарных проверок

static long long      all_checks = 0; // число элементарных проверок
static u_long	Nintr = 0;	// количество полученных прерываний
static char	*str_excl = NULL;
static char	*prog_send = NULL; // имя "подыгрывающей" программы
static char	*prog_send2 = NULL; // имя "подыгрывающей" программы N 2
static int	NpinP = -1;	// N   пина для "подыгрывающей" программы
static int	NpinP2 = -1;	// N 2 пина для "подыгрывающей" программы
static int	nexcl = 0;	// число исключенных пинов из тестирования

static char *fname = NULL;

static char cmdline[1024] = { '\0', };

#define PIN_SOLDER(N)		pins[N].solder		// пин запаян
#define PIN_AVAIL(N)		pins[N].avail		// пин exported/
							//	unexported
#define PIN_NEVER_EXPORT(N)	pins[N].never_export	// исключен из проверок согласно парам
#define PIN_NEVER(N)		pins[N].never		// исключен из проверок
#define PIN_ACTIVE_LOW(N)	pins[N].active_low	// "0" - signal is "0",
							// "1" - signal is "1"
#define PIN_NO_OPEN(N)		pins[N].no_open		// экспортровать не
							//	удалось
#define PIN_DIRECTION(N)	pins[N].direction	// =-1 - не определено, =0 - IN, =1 - OUT

#define EXIT_FAIL_MSG(NM)	{					\
	print_msg("exit_fail_msg :errors =%d, limit_error =%d. %s.%d\n",\
		errors, limit_error, __func__, __LINE__);		\
	exit_fail_msg(NM);						\
}

#define EXIT_PASS_MSG(NM)	{					\
	print_msg("exit_pass_msg :errors =%d, limit_error =%d\n",	\
		errors, limit_error);					\
	exit_pass_msg(NM);						\
}

#ifndef E90_linux

typedef  long long  hrtime_t;

#define	print_msg(fmt, args...)		\
			printf( fmt "", ## args)
#define	print_msg_error(fmt, args...)	\
			printf( "OШИБKA :"fmt "", ## args)
#define	print_msg_warning(fmt, args...)	\
			printf( "ПPEДУПPEЖДEHИE :"fmt "", ## args)
#define	print_msg_note(fmt, args...)	\
			printf( "ЗАМЕЧАНИЕ :"fmt "", ## args)

#define	exit_pass_msg(fmt, args...)	\
	    do{\
		printf("ФИНИШ_Т: УСПЕХ: " fmt "\n", ## args);\
		exit( 0);}while(0)

#define	exit_fail_msg(fmt, args...)	\
	    do{\
		printf("ФИНИШ_Т: ОТКАЗ: " fmt "\n", ## args);\
		exit( -1);} while(0)

#define start_t(fmt, args...) print_msg("СТАРТ_Т: "   fmt, ## args "\n");

/*
static inline hrtime_t gethrtime (void)
{
    hrtime_t usec_time;
#ifdef USE_GETTIMEOFDAY
   static struct timeval time;
   gettimeofday(&time, 0);
   usec_time = (HRT_SEC(time.tv_sec) + time.tv_usec);
#else  // no USE_GETTIMEOFDAY
   long rval;
   rval = el_get_usec(&usec_time);
   if (rval != 0) {
        return 0;
   }
#endif // USE_GETTIMEOFDAY
  return(usec_time);
} // gethrtime
*/

#else /*E90_linux*/

#include <linux/msg_tst.h>
#include <linux/tst_includes.h>
#include <linux/gethrtime.h>

#endif /*E90_linux*/

static hrtime_t
Gethrtime (void) {
    struct timespec tv;
    hrtime_t tval;

    clock_gettime(CLOCK_REALTIME, &tv);
    tval = (HRT_SEC(tv.tv_sec) + (hrtime_t)tv.tv_nsec);
    return (tval);
}

static void alarm_(int, char *);
static void sleep_mlsec(int);
static hrtime_t test_run_time, t_start_tst, t_start_prog;

static int     *ranges = NULL;
static int     nrange = -1;

static int
parse_str_excl (char *sp_par) {
        int sret, i;
        char* sp = sp_par;
        char* sp_ = sp;
        char* spend = sp_par + strlen(sp_par);

        ranges = (int *)malloc(NGPIO * sizeof(int));

        if (vverbose)
                print_msg("Разбор параметра EXCL =%s\n", sp_par);

        for (nrange = 0; ;) {
                int b, e;
                int ism, isd;

                if (sp >= spend)
                        break;
                if (isspace(sp[0])) {
                        sp++;
                        continue;
                }
                if (vverbose)
                        print_msg("%d :%s\n", nrange, sp);
                b = e = NO_CNCT;
                sret = sscanf(sp,"%d-%d", &b, &e);
                if (vvverbose)
                        print_msg("%s `%d`-`%d`, sret =%X, sp =%s\n",
                                "%d-%d", b, e, sret, sp);
                isd = 0;
                if (sret == 2 && b != NO_CNCT && e != NO_CNCT) {
                        isd = 1;
                }
                if (sret == 1 && b != NO_CNCT && b < 0 && e == NO_CNCT) {
                        e = -b;
                        b = 0;
                        isd = 1;
                }
                ism = 0;
                sp_ = sp;
                for (i = 0; ;i++) {
                        if (vvverbose)
                                print_msg("%d Разбор sp :sp_ =%s, sp =%s\n",
                                        i, sp_, sp);
                        if (isdigit(sp[0]))
                                sp++;
                        else if (sp[0] == '-' && !ism) {
                                sp++;
                                ism = 1;
                        }
                        else if (isspace(sp[0]) || (sp[0] == ',')) {
                                sp++;
                                break;
                        }
                        else if (sp < spend) {
                    	    print_msg_error("%s '%s' :\n\tнедопустимое "
				"значение =%s !\n", FNC, sp_par, sp);
                                return 1;
                        }
                        else
                                break;
                }

                if ((sp_ == sp) || (sp[0] == ',')) {
                        sp++;
                        continue;
                }
                if (e == NO_CNCT && ism)
                        e = NGPIO -1;
                else if (e == NO_CNCT)
                        e = b;

                if (vvverbose)
                        print_msg("%d\t%d %d, sp =%s\n", nrange, b, e, sp);
                if (vverbose) {
                        if (b == e)
                                print_msg("\t%2d)\t%2d\n", nrange, b);
                        else
                                print_msg("\t%2d)\t%2d - %2d\n", nrange, b, e);
                }
                ranges[nrange++] = b*1000 +e;

                if (nrange >= NGPIO) {
                        print_msg_warning("parse_str_excl :"
                                "слишком много задано диапазонов >NGPIO =%d\n",
                                NGPIO);
                        break;
                }
        }

        /*
         * Печать таблицы
         * диапазонов исключаемых пин-ов по ключу EXCL.
         */
        if (verbose) {
                char msg[256], *msg_ = msg;
                for (i = 0; i < nrange; i++) {
                        int d, b, e;
                        d = ranges[i];
                        b = d / 1000;
                        e = d % 1000;
                        if (b == e)
                                msg_ += sprintf(msg_, "\t%2d)\t%2d\n", i, b);
                        else
                                msg_ += sprintf(msg_, "\t%2d)\t%2d - %2d\n",
                                                i, b, e);
                }
                print_msg("Таблица диапазонов\n\t"
                        "исключаемых пин-ов по ключу EXCL :%s\n%s",
                        sp_par, msg);
        }
        return 0;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ parse_str_excl */

/*
 *   Инициализация таблицы линий связи pins
 * по таблице ranges - диапазонов исключаемых пин-ов EXCL.
 */
static int
parse_excl_pins () {
#define IN_RANGE(N)     ((N >= 0) && (N < NGPIO))
    int     i, res = 0;

    if (vverbose)
	print_msg("Инициализация таблицы линий связи, "
            "число диапазонов =%d\n", nrange);
    if (nrange > 0) {
	int d, b, e, n;
        for (i = 0; i < nrange; i++) {
    	    d = ranges[i];
            b = d / 1000;
            e = d % 1000;

            if (vvverbose)
        	if (b == e)
            	    print_msg("\t%2d)\t%2d\n", i, b);
                else
            	    print_msg("\t%2d)\t%2d - %2d\n", i, b, e);

	    if (!IN_RANGE(b) || !IN_RANGE(e)) {
        	if (b != e) {
		    if ((b < NGPIO) && (e > NGPIO)) {
            		print_msg_note("\tдиапазон N %d (%2d - %2d) неверный\n"
			    "\t\t%2d -> %2d\n",
			    i, b, e, e, NGPIO -1);
			e = NGPIO -1;
		    }
		    else {
			res++;
            		print_msg_error("\tдиапазон N %d (%2d - %2d) "
                	    "неверный !!!\n", i, b, e);
			continue;
		    }
		}
                else {
		    res++;
            	    print_msg_error("\t%d - пин N %d неверный !!!\n", i, b);
                    continue;
		}
    	    }
            for (n = b; n <= e; n++) {
        	if (PIN_NEVER_EXPORT(n)) {
            	    print_msg_note("\t%d :пин N %d уже исключен\n", i, n);
                    continue;
                }
                PIN_NEVER(n) = PIN_NEVER_EXPORT(n) = 1;
                nexcl++;
    	    }
	}
    }

    if (nrange)
	free(ranges);
    return res;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ parse_excl_pins */

/*
 * Определить число пинов gpio по sysfs-интерфейсу.
 */
static int
define_ngpio () {
	char	nmcmd[64];
	FILE	*fp;
	char	cmd[128];
	char	fngpio[64];
	int	fdf;
	int	res;

	sprintf(nmcmd, "/tmp/cmd.%d", my_pid);
	fp = fopen(nmcmd, "w");
    	sprintf(fngpio, "/tmp/ngpio.%d", my_pid);
	sprintf(cmd, "%sS=0\nfor N in `cat %s/gpiochip*/ngpio`\n"
		"do ((S+=N)); done\necho \"$S \" > %s",
		vverbose ? "set -x\n" : "", DGPIO, fngpio);
	fwrite(cmd, sizeof(char), strlen(cmd), fp);
	fclose(fp);
	sprintf(cmd, "/bin/bash %s; rm -rf %s", nmcmd, nmcmd);
	if (vverbose)
		print_msg("%s :cmd =`cat %s`\n", FNC, nmcmd);
	system(cmd);
	if (!access(fngpio, F_OK)) {
		fdf = open(fngpio, O_RDONLY, 0);
		if (fdf < 0) {
			if (vverbose)
				print_msg_warning("%s :файл не открылся: %s\n",
					fngpio, strerror(errno));
			return -1;
		}

		res = read(fdf, cmd, 3);
		res = atoi(cmd);
		close(fdf);
		unlink(fngpio);
	}
	else {
		print_msg_warning("access %s :нет файла.\n", fngpio);
			return -1;
	}
	if (verbose)
	    print_msg("%s, res =%d\n", FNC, res);
	return res;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ define_ngpio */


static int	fexport = -1;
static int	funexport = -1;
static char export_name[MAX_SYSFS_NAME_LEN] = { '?' };
static char unexport_name[MAX_SYSFS_NAME_LEN] = { '?' };

static int
gpio_pin_open (int pin) {
    int	szNpin, rwrite;
    char Npin[MAX_COMMAND_LEN];
    char gpio_name[MAX_SYSFS_NAME_LEN];

    if (PIN_NEVER(pin))
	return FILE_ERR;

    sprintf(gpio_name, "%s/%s%d", DGPIO, pins_name, pin);
    if (fexport == -1) {
        sprintf(export_name, "%s/export", DGPIO);
        fexport = open(export_name, O_WRONLY);

        if (fexport < 0) {
    	    exit_fail_msg("%s :!!! Ошибка открытия %s - %m\n",
		FNC, export_name);
        }
	if (verbose)
	    print_msg("%s(%d) :Файл = %s - открыт, Nfd =%d\n",
		FNC, pin, export_name, fexport);
    }

    sprintf(Npin, "%d", pin);
    rwrite = write(fexport, Npin, szNpin = strlen(Npin));
    if (rwrite != szNpin) {
	print_msg_error("Ошибка export-ирования %s - %m\n", gpio_name);
	print_msg_error("%s(%d) :rwrite =%d != szNpin =%d\n",
	    FNC, pin, rwrite, szNpin);
    }

    /*
     * Проверка export-ирования линии GPIO
     */
    if (!access(gpio_name , F_OK)) { /* появился файл gpio<пин> */
	if (verbose)
	    print_msg("линия GPIO =%s\texport-ирована\n", gpio_name);
    } else {
	print_msg_error("Нет линии = %s\n", gpio_name);
	PIN_NEVER(pin) = PIN_NO_OPEN(pin) = 1;
	return FILE_ERR;
    }
    return 0;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_open */

static void
gpio_pin_close (int pin) {
    int szNpin, rwrite;
    char Npin[MAX_COMMAND_LEN];
    char gpio_name[MAX_SYSFS_NAME_LEN];

    if (!PIN_AVAIL(pin))
	return;

    if (funexport == -1) {
        sprintf(unexport_name, "%s/unexport", DGPIO);
        funexport = open(unexport_name, O_WRONLY);
        if (funexport < 0) {
            exit_fail_msg("%s :!!! Ошибка открытия %s - %m\n",
                FNC, unexport_name);
        }
        if (verbose)
	    print_msg("%s(%d) :Файл = %s - открыт, Nfd =%d\n",
        	FNC, pin, unexport_name, funexport);
    }

    sprintf(Npin, "%d", pin);
    rwrite = write(funexport, Npin, szNpin = strlen(Npin));
    if (rwrite != szNpin) {
        print_msg_error("Ошибка un-export-ирования %s - %m\n", gpio_name);
        print_msg_error("%s(%d) :rwrite =%d != szNpin =%d\n",
            FNC, pin, rwrite, szNpin);
    }

/*
 * Проверка un-export-ирования линии GPIO
 */
    sprintf(gpio_name, "%s/%s%d", DGPIO, pins_name, pin);
    if (!access(gpio_name , F_OK)) { /* не удалился файл gpio<пин> */
        print_msg_error("!!! линия = %s НЕ un-export-ирована !!!\n", gpio_name);
    } else if (verbose)
            print_msg("линия GPIO =%s\tun-export-ирована\n", gpio_name);
    return;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_close */

static int
gpio_pin_open_val (int pin) {

	char val_name[MAX_SYSFS_NAME_LEN];
	int fd;

	if (PIN_NO_OPEN(pin))
		return FILE_ERR;

	sprintf(val_name, "%s/gpio%d/value", DGPIO, pin);

	fd = open(val_name, O_RDWR | O_NONBLOCK);
	if (verbose)
		print_msg("open(%s) =%d\n", val_name, fd);
	if (fd < 0) {
		print_msg_error("%s :Ошибка открытия %s %s\n",
			FNC, val_name, strerror(errno));
		return FILE_ERR;
	}
	return fd;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_open_val */

static int
gpio_pin_set_direction (int pin, int direction) {
    char direction_name[MAX_SYSFS_NAME_LEN];
    char command[MAX_COMMAND_LEN];

    FILE *fdir = NULL;

    if (PIN_NO_OPEN(pin)) {
	if (vvverbose)
	    printf("\t%s :return FILE_ERR - PIN_NO_OPEN !!!\n", FNC);
	return FILE_ERR;
    }

    if ((verbose > 2) && (PIN_DIRECTION(pin) == direction)) {
	if (vvverbose)
	    print_msg("set_fdir\tуже установлено\t\t=%s\n",
		direction == DIRECTION_OUT ? "out" : "in");
	    return 0;
    }

    sprintf(direction_name, "%s/%s%d/direction", DGPIO, pins_name, pin);
    fdir = fopen(direction_name, "w");
    if (!fdir) {
	print_msg_error("%s :Ошибка открытия %s\n", FNC, direction_name);
	PIN_NEVER(pin) = PIN_NO_OPEN(pin) = 1;
	if (vvverbose)
	    printf("\t%s :return FILE_ERR\n", FNC);
	return FILE_ERR;
    }

    if (direction == DIRECTION_OUT)
	sprintf(command, "out");
    else
	if (direction == DIRECTION_IN)
	    sprintf(command, "in");
	else {
	    fclose(fdir);
	    if (vvverbose)
		printf("\t%s :return FILE_ERR\n", FNC);
	    return INPUT_ERR;
	}

    if (vvverbose) {
	sprintf(direction_name, "DGPIO/%s%d/direction", pins_name, pin);
	print_msg("set_direction\t%s\t=%s\n", direction_name, command);
    }

    fputs(command, fdir);
    fclose(fdir);
    PIN_DIRECTION(pin) = direction;

    return 0;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_set_direction */

static int
gpio_pin_set_level (int pin, int level) {
	char value_name[MAX_SYSFS_NAME_LEN];
	char *command;

	FILE *value = NULL;

	if (PIN_NO_OPEN(pin))
		return FILE_ERR;

	sprintf(value_name, "%s/%s%d/value", DGPIO, pins_name, pin);

	value = fopen(value_name, "w");
	if (!value) {
		print_msg_error("%s :Ошибка открытия %s\n", FNC, value_name);
		PIN_NEVER(pin) = PIN_NO_OPEN(pin) = 1;
		return FILE_ERR;
	}

	if (level == LEVEL_LOW)
		command = "0";
	else if (level == LEVEL_HIGH)
		command = "1";
	else {
		fclose(value);
		return INPUT_ERR;
	}

	if (vvverbose) {
		sprintf(value_name, "SYSFS/%s%d/value", pins_name, pin);
		print_msg("set_level\t\t%s\t=%s\n", value_name, command);
	}

	fputs(command, value);
	fclose(value);

	return 0;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_set_level */

static int
gpio_pin_get_value (int pin) {

	char value_name[MAX_SYSFS_NAME_LEN];
	char str[MAX_COMMAND_LEN];

	FILE *value = NULL;

	if (PIN_NO_OPEN(pin))
		return FILE_ERR;

	sprintf(value_name, "%s/%s%d/value", DGPIO, pins_name, pin);

	value = fopen(value_name, "r");
	if (!value) {
		print_msg_error("%s :Ошибка открытия %s\n", FNC, value_name);
		PIN_NEVER(pin) = PIN_NO_OPEN(pin) = 1;
		return FILE_ERR;
	}

	fgets(str, 2, value);
	fclose(value);

	if (vvverbose) {
		sprintf(value_name, "DGPIO/%s%d/value", pins_name, pin);
		print_msg("%s\t\t%s\t@%s\n", FNC, value_name, str);
	}
	return atoi(str);
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_get_value */

static int
gpio_pin_get_active_low (int pin) {

	char active_low_name[MAX_SYSFS_NAME_LEN];
	char str[MAX_COMMAND_LEN];

	FILE *active_low = NULL;

	if (PIN_NO_OPEN(pin))
		return FILE_ERR;

	sprintf(active_low_name, "%s/%s%d/active_low",
		DGPIO, pins_name, pin);

	active_low = fopen(active_low_name, "r");
	if (!active_low) {
		print_msg_error("%s Ошибка открытия %s\n",
		    FNC, active_low_name);
		PIN_NEVER(pin) = PIN_NO_OPEN(pin) = 1;
		return FILE_ERR;
	}

	fgets(str, 2, active_low);
	fclose(active_low);

	if (vvverbose) {
		sprintf(active_low_name, "DGPIO/%s%d/active_low",
			pins_name, pin);
		print_msg("get_active_low\t%s\t@%s\n", active_low_name, str);
	}

	return atoi(str);
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_get_active_low */

static int
gpio_pin_set_edge (int pin, char *edge) {

	char name_edge[MAX_SYSFS_NAME_LEN];

	FILE *file_edge = NULL;

	if (PIN_NO_OPEN(pin))
		return FILE_ERR;

	sprintf(name_edge, "%s/%s%d/edge", DGPIO, pins_name, pin);

	file_edge = fopen(name_edge, "w");
	if (!file_edge) {
		print_msg_error("%s, Ошибка открытия %s - %s\n",
			FNC, name_edge, strerror(errno));
		return FILE_ERR;
	}

	fputs(edge, file_edge);
	fclose(file_edge);

	if (vvverbose) {
		sprintf(name_edge, "DGPIO/%s%d/edge",
			pins_name, pin);
		print_msg("set_edge\t\t%s\t=%s\n", name_edge, edge);
	}

	return 0;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_set_edge */

//ZZZ - ??? - зачем эта функция ???
static int
pin_is_valid (int pin) {
	int res = 1;

	if (PIN_NEVER_EXPORT(pin))
		res = 0;

	if (vvverbose)
		print_msg("%2d pin_is_valid\t=%d\n", pin, res);

	return res;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ pin_is_valid */


/*
 *    Проверка пин-ов, подлежащих проверкам,
 * на сменяемость уровня сигнала.
 * Eсли уровень не изменяется, то трактуем, что он 'запаян'.
 */
static void
check_solder () {
	int i;
	int value;
	int signal;

	/*
	 * установить всем OUT-пинам низкий уровень :
		set_direction "out"
		set_level PIN_ACTIVE_LOW(<Nпина>));
	 */
	if (vvverbose)
		print_msg_note("%s :"
			"установить всем OUT-пинам низкий уровень\n", FNC);

	for (i = 0; i < NGPIO; i++) {
		if (PIN_NEVER(i))
			continue;
		gpio_pin_set_direction(i, DIRECTION_OUT);
		gpio_pin_set_level(i, PIN_ACTIVE_LOW(i));
		
	}
	if (without_check_solder)
		return;

	/*
	 * считать уровени на IN-пинах.
	 * Если он высокий - считать, что пин 'запаян'.
	 */
	if (vvverbose)
		print_msg_note("%s :считать уровени на IN-пинах\n", FNC);
	for (i = 0; i < NGPIO; i++) {
		if (PIN_NEVER(i))
			continue;

		signal = PIN_ACTIVE_LOW(i);
		gpio_pin_set_direction(i, DIRECTION_IN);
		value = gpio_pin_get_value(i);
		if (signal != value) {
			print_msg_note("Пин %2d 'запаян' => %d, "
			    	"исключается из проверок\n", i, value);
			PIN_NEVER(i) = PIN_SOLDER(i) = 1;
			continue;
		}
		gpio_pin_set_direction(i, DIRECTION_OUT);
		gpio_pin_set_level(i, PIN_ACTIVE_LOW(i));
	}
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ check_solder */


static void
prn_state () {
	int	i;
	char msg[512], *msg_ = msg;

	msg_ += sprintf(msg_, "\nprn_state :Состояние пинов\n"
		"(No_Open, Solder, Avail, Never_Export, NEver, Active_Low)\n"
			"\tNO S A NE N AL\n");
	for (i = 0; i < NGPIO; i++) {
		msg_ += sprintf(msg_, "%2d)\t%d  %d %d %d  %d ", i,
				PIN_NO_OPEN(i),
				PIN_SOLDER(i),
				PIN_AVAIL(i),
				PIN_NEVER_EXPORT(i),
				PIN_NEVER(i));
		if (PIN_ACTIVE_LOW(i) >=0)
			msg_ += sprintf(msg_, "%d\n", PIN_ACTIVE_LOW(i));
		else
			msg_ += sprintf(msg_, "-\n");
	}
	print_msg("%s\n", msg);

} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ prn_state */


static int
init_poll () {
    int	res = 0, n, N = 0;

    for (n = 0; n < NGPIO; n++)
	if (!PIN_NEVER(n))
	    NPOLFD++;

    print_msg("~~~\t%s :NGPIO =%d, NfdPOLL =%d ~~~~~~~\n", FNC, NGPIO, NPOLFD);

    pollfd = (struct pollfd *)malloc(NPOLFD * sizeof(struct pollfd));

//  таблица для перевода N пина в N pollfd
    pin2poll = (int *)malloc(NGPIO * sizeof(int));

//  таблица для перевода N pollfd в N пина
    poll2pin = (int *)malloc(NPOLFD * sizeof(int));

//  таблица для хранения fd пин-файлов
    pin2fd = (int *)malloc(NGPIO * sizeof(int));

    for (n = 0, N = 0; n < NGPIO; n++) {
	if (PIN_NEVER(n)) {
	    pin2fd[n] =
	    pin2poll[n] = -1;
	    continue;
	}
	pin2fd[n] = gpio_pin_open_val(n);
	poll2pin[N] = n;
	pin2poll[n] = N++;

	if (vverbose)
	    printf("N pin =%2d,\tNpoll =%2d -> %2d\tfd =%2d\n",
		n, N-1, poll2pin[N-1], pin2fd[n]);
    }
    if (verbose)
	print_msg("%s :res =%d\n", FNC, res);
    return res;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ init_poll */

static struct timespec slp_time;

static void
sleep_mlsec (int time_mlsec) {
	slp_time.tv_sec = time_mlsec/1000;
	slp_time.tv_nsec = (time_mlsec%1000) *1000000;
	nanosleep(&slp_time, NULL);
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ sleep_mlsec */


static void
READ_FD (int NPIN, char *MSG)	{
	char buf[16] = {0};
	if (vverbose)
	    printf("READ_FD(%d, %s)\n", NPIN, MSG);
	read(pin2fd[NPIN], buf, sizeof(buf));
	if (verbose && buf[0])
		print_msg("%s :%2d read =%s", MSG, NPIN, buf);
}

static void
prn_pollfd (char *msg) {
    int J;

    for (J = 0; J < NPOLFD; J++)
	printf("%2d)\tpin =%2d\tfd =%2d, events =0X%02X, revents =0X%02X\n",
	    J, poll2pin[J], pollfd[J].fd, pollfd[J].events, pollfd[J].revents);
}

/*
    Работа с прерываниями :
Через интерфейс sysfs можно запросить прерывания по изменению состояния пина.
Установка прерывания производится путём записи значения в файл "edge".
    (edge - край,кромка,грань т.е. = фронт)
Значения могут быть:
    - none	- отключить прерывание;
    - rising	- включить прерывание по нисходящему фронту;
    - falling	- включить прерывание по восходящему фронту;
    - both 	- включить прерывание по обеим фронтам.
 */

static int
wait_inters () {	// ожидание прерываний	- waiting for interrupts
/*
 * Все проверяемые пины настроим на прием прерывания
 * по :
    wait_regime = 1 - переднему фронту "rising";
    wait_regime = 2 - заднему   фронту "falling";
    wait_regime > 2 - по обеим фронтам "both"
 */
    int err = 0, i, j, J;
    char *sedge, *medge;
    char mch[64], mrevents[128], *mr;
    int poll_res;

    if (!wait_regime)
	wait_regime = 1;
    sedge = wait_regime == 1 ? "rising" :
		wait_regime == 2 ? "falling" : "both";
    medge = wait_regime == 1 ? "переднему" :
		wait_regime == 2 ? "заднему" : "обоим";

    print_msg_note("=== ожидания прер. по фронту '%s' (edge =%s) ===\n",
	    medge, sedge);

    for (i = 0; i < NGPIO; i++) {
        if (pin2fd[i] < 0)
            continue;
        gpio_pin_set_direction(i, DIRECTION_IN);
        gpio_pin_set_edge(i, sedge);
        READ_FD(i, sedge);
    }
    /*
     *  создаем таблицу pollfd
     */
    for (i = 0, J = 0; i < NGPIO; i++) {
	if (pin2fd[i] < 0)
	    continue;
	pollfd[J].fd = pin2fd[i];
        pollfd[J].events = POLL_CH;
        pollfd[J].revents = 0;
        pin2poll[i] = J++;
    }

    t_start_tst = Gethrtime();
    if (prog_send) {	// !!! для ОТЛАДКИ - запуск скрипта генерации прер.
	char cmd[256], sNP[32] = {' '}, sNP2[32] = {' '};

	if (NpinP >=0)
	    sprintf(sNP, "%d", NpinP);
	if (NpinP2 >=0)
	    sprintf(sNP2, "%d", NpinP2);
	if (prog_send2)
	    sprintf(cmd, "export PATH=.:$PATH; %s %lld %s & %s %lld %s &",
		prog_send, t_start_tst/1000000LL, sNP,
		prog_send2, t_start_tst/1000000LL, sNP2);
	else
	    sprintf(cmd, "export PATH=.:$PATH; %s %lld %s &",
		prog_send, t_start_tst/1000000LL, sNP);
	if (verbose)
	    print_msg_note("\n\tsystem('%s')\n", cmd);
	system(cmd);
    }

    for (all_checks = 0;
	(number_checks < 0) || (all_checks < number_checks);
	all_checks++
	) {
        int Nrevents, revents, time_out = 10000	/* mlsec */;

	for (i = 0, J = 0; i < NGPIO; i++) {
	    if (pin2fd[i] < 0)
		continue;
    	    pollfd[J].events = POLL_CH;
    	    pollfd[J++].revents = 0;
	}
    /*
     *	Через poll по всем допустимым пинам, ожидать появления прерываний.
     */
        if (vverbose)
            prn_pollfd("до    poll");
        alarm_(11, "Ожидание прерываний (poll)");
        errno = 0;
    /*
     * reset_irqs :
     *  перед выдачей сигнала на проверяемом пине,
     *  надо выполнить чтение с .../gpio<пин>/value.
     */
        for (J = 0; J < NPOLFD; J++) {
	    j = poll2pin[J];
            if (vvverbose)
                print_msg("reset_irqs :%d read\n", j);
            READ_FD(j, "reset_irqs");
        }

	mr = mrevents;
	if (timeout_before_check)
	    sleep_mlsec(timeout_before_check);

        poll_res = poll(pollfd, NPOLFD, time_out);  // ож. 10 сек.
	test_run_time = Gethrtime() - t_start_tst;
	sprintf(mch, "%5.3f cек:\t- Проверка N %lld, ",
	    (float)(test_run_time)/HRT_SEC(1), all_checks +1);

	Nrevents = 0;
        for (J = 0; J < NPOLFD; J++)
    	    Nrevents += pollfd[J].revents > 0;
        if (verbose) {
	    char msg[64];
	    sprintf(msg, "после poll, poll_res =%d, Nrevents =%d",
		poll_res, Nrevents);
            prn_pollfd(msg);
	}

        if (!poll_res) {
	    char merr[128] = { ' ' };
	    if (errno)
		sprintf(merr, " - %s", strerror(errno));
    	    print_msg_note("%stime_out =%d mlsec ожидания прер. по фронту =%s%s\n",
		    mch, time_out, medge, merr);
        }

        if (poll_res < 0) {
            err = 1;
            print_msg_error("%sожидания прер. по %s - %m\n", mch, sedge);
            goto end_check;
        }

        for (J = 0; j = poll2pin[J], J < NPOLFD; J++) {
	    int	value;
	    if (wait_regime == 3)
		value = gpio_pin_get_value(j);
            if (revents = pollfd[J].revents) {
		mr += sprintf(mr, " %d%s", poll2pin[J],
			wait_regime == 3 ? value ? "+" : "-" : "");
                if ((revents & POLLERR) && errno)
                    if (errpoll++ < limit_errpoll)
                        if (vverbose)
                            print_msg("%srevents N %2d =0X%02X%s\n",
                                mch, j, revents,
                                errpoll == limit_errpoll ? " ..." : "");
                if (verbose)
                    if (revents & POLL_CH)
                        print_msg("%sесть прер. у N %2d%s, "
                            "revents =0X%02X\n", mch, j,
			    wait_regime == 3 ? value ? " =+" : " =-" : "",
			    revents);
                    else
                        print_msg("%srevents у N %2d =0X%02X\n",
                            mch, j, revents);
                if ((revents & POLL_CH) == 0)
                    continue;
                Nintr++;
                READ_FD(j, "is revents");
            }
        } /* end for ................................ `Проверки poll` */

	if (!verbose && (mrevents != mr))
	    printf("%s есть события у %s\n", mch, mrevents);

	if (wait_time) {
            if (test_run_time > HRT_SEC(wait_time))
		break;
	}
    }
end_check :
    return err;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ wait_inters */


/*
 * Устанавливает обработчик @hndlr на сигнал @sig_num
 */
static void
set_sig_handler (int sig_num, void (*hndlr)(int, siginfo_t *, void *)) {
	struct sigaction act;

	memset (&act, '\0', sizeof(act));

	act.sa_sigaction = hndlr;
	act.sa_flags = SA_SIGINFO;

	if (sigaction(sig_num, &act, NULL) < 0) {
		perror ("sigaction");
		return;
	}
}

static char	*msg_alarm = "";
static int	t_alarm = 0;

static void
alarm_ (int time, char *msg) {
	msg_alarm = msg;

	if (vvverbose)
		print_msg("alarm_ :%s, set time =%d sec\n", msg, time);
	alarm(t_alarm = time);
	return;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ alarm_ */


static void
close_exit () {
    int i, fd = 0;
    char msg[4096], *msg_ = msg;
/*
 *  Закрытие всех ранее
 *  export-ированных пинов GPIO и всех ранее открытых файлов pollfd.
 */
    for (i = 0; i < NGPIO; i++) {
	gpio_pin_close(i);
	if (pin2fd != NULL) {
	    fd = pin2fd[i];
	    if (fd > 0)
		close(fd);
	}
	if (vverbose && (fd > 0))
	    msg_ += sprintf(msg_, "\tN pin =%d\tfd =%d\n", i, fd);
	else if (vvverbose)
	    msg_ += sprintf(msg_, "\tN pin =%d\n", i);
    }
    if (msg != msg_) {
	print_msg("%s :\n%s", FNC, msg);
	prn_state();
    }
    if (fexport > 0)
	close(fexport);
    if (funexport > 0)
	close(funexport);

    if (errpoll)
	print_msg_note("Количество полученных прерываний =%ld, "
	    "N ошибок poll =%d\n", Nintr, errpoll);
    else
	print_msg("Количество полученных прерываний =%ld\n", Nintr);
	
    print_msg("Время работы :теста =%5.3f, программы =%5.3f сек.\n",
	(float)(Gethrtime() - t_start_tst)/HRT_SEC(1),
	(float)(Gethrtime() - t_start_prog)/HRT_SEC(1));
    if (errors)
	EXIT_FAIL_MSG(NM_TST);
    exit_pass_msg(NM_TST);
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ close_exit */

/*
 * Обработчик сигналов SIGINT - <Control+C>, SIGALRM;
 */
static void
sigcommon_hndlr (int sig, siginfo_t *siginfo, void *context) {
	switch ( siginfo->si_signo ) {
	case SIGINT:
		print_msg_note("Тест прерван пользователем\n");
		break;
	case SIGALRM:
		if (t_alarm)
			print_msg_note("SIGALRM. Тест прерван."
				"\n\tСработал таймер на '%s' =%d сек.\n",
				msg_alarm, t_alarm);
		break;
	default:
		print_msg("%s(): dummy handler for signal [%d]\n",
			__func__, siginfo->si_signo);
	}

	close_exit();
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ sigcommon_hndlr */


static int usage_and_exit = -1;

#define usageErr "\n\
Instructions for using the %s :\n\
       %s -h|H\n\
   ,where:\n\
	-h	- print usage in english;\n\
	-H	- print usage in russian.\n\
"
#define usageR "\n\
Инструкция по использованию программы %s:\n\
       %s -h|H\n\
    ,где:\n\
          -h    - print usage (на английском);\n\
          -H    - печать данной инструкции.\n\
    или\n\
       %s [-N <max Ngpio>][-t <время>]|[-C <число проверок>]\n\
	    [-w|ww|www] [-e <список пинов>] [-p <имя прог.]\n\
    ,где:\n\
	  -N <max Ngpio>    - число пинов gpio;\n\
	  -p <name prog.>   - имя \"подыгрывающей\" программы;\n\
	  -e <список пинов> - задание пинов исключенных из тестирования\n\
	     <список пинов> - \"список пинов\" | 'список пинов'\n\
	      список пинов  - [N пина|диапазон пинов] (через пробел|запятую)\n\
\n\
          -t <время> - время тестирования в секундах\n\
		(по умолчанию =до прерывания тестирования сигналом <Ctrl>+C)\n\
	  -t|-C - тест завершится по событию, 'что раньше'.\n\
	  -w|ww|www - режим ожидания прерываний :\n\
                w   - по переднему фронту  (rising);\n\
                ww  - по заднему   фронту  (falling);\n\
                www - по обеим     фронтам (both)\n\
\n\
  ДОПОЛНИТЕЛЬНЫЕ параметры :\n\
       -l <число ошибок> - лимит ошибок (по умочанию =%d);\n\
       -L <число ошибок> - лимит отображаемых сообщенийй с ошибами errpoll\n\
			  (по умочанию =%d);\n\
       -s - без контроля запаяных пинов (для отладки);\n\
       -S <время милисек> - задержка перед ожиданием прерываний;\n\
       -v|vv|vvv - печать дополнительной информации.\n\
  Параметр разработчика :-d - dbgopt\n\
\n\
  Примеры <списков пинов> :\n\
    -e \"0-5, 15-999\"	# исключить пины 0-5 и все от 15 - до конца;\n\
    -e ' -5, 15-'	# исключить пины 0-5 и все от 15 - до конца;\n\
			# !!! в списке первый символ не может быть знаку ='-'\n\
    -e '5 7,29-'	# исключить пины 5, 7, и 29 - до конца;\n\
\n\
"
#define usageE "\n\
Instructions for using the %s :\n\
       %s -h|H\n\
   ,where:\n\
	  -h	- print usage in english;\n\
	  -H	- print usage in russian.\n\
    or\n\
	%s [-N <max Ngpio>][-t time in seconds]|[-C <number of checks>]\n\
	    [-w|ww|www] [-e <pin list>] [-p <name prog.>]\n\
   ,where:\n\
          -N <max Ngpio>    - number of gpio pins;\n\
	  -p <name prog.> - name of the \"playing\" program;\n\
          -t <time> - testing time in seconds\n\
		(default =before test is interrupted by <Ctrl>+C);\n\
	  -t|-C - test will come to end on 'that earlier'.\n\
	  -e <pin list> - setting pins excluded from testing\n\
	     <pin list> - \"pin list\" | 'pin list'\n\
	      pin list  - [N pin|pin range] (separated by spaces or commas)\n\
\n\
	  -w|ww|www - interrupt wait mode :\n\
	   w   - on the front (rising);\n\
	   ww  - on the back front (falling);\n\
	   www - on both fronts (both)\n\
   ,where:\n\
       -c - CHECK connections by <file-param>;\n\
       -cc + -f <file-param> - CHECK connections by <file-param>\n\
                              +CHECK all blankings;\n\
       -cc - definition of plugs connections\n\
             and creating <file-param> =auto-param.\n\
          This check 'requires' the operator to specify a list of pins\n\
          to be excluded from testing.\n\
  ADDITIONAL parametrs :\n\
       -l <number of errors> - limit of errors ошибок (by default =%d);\n\
       -L <number of errors> - limit of displayed errpoll messages\n\
			      (by default =%d);\n\
       -s - without monitoring of soldered pins (for debugging);\n\
       -v|vv|vvv - printing additional information\n\
  Developer option :-d - dbgopt\n\
\n\
   Examples of <pin lists> :\n\
    -e \"0-5, 15-999\"	# exclude pins 0-5 and everything from 15 to the end;\n\
    -e '-5, 15-' 	# exclude pins 0-5 and everything from 15 to the end;\n\
			# in list, first character cannot be ='-'\n\
    -e '5 7,29-'	# eliminate pins 5, 7, and 29 - to the end;\n\
\n\
"

static void
usage (int arg) {
    if (arg == 0) {
	printf(usageE, my_prog_name, my_prog_name, my_prog_name,
	    limit_error, limit_errpoll);
	return;
    }
    if (arg == 1) {
	printf(usageR, my_prog_name, my_prog_name, my_prog_name,
	    limit_error, limit_errpoll);
	return;
    }

    printf(usageErr, my_prog_name, my_prog_name);
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ usage */

static int
set_current_dir (char *prog) {
	int	i, n, res = 0, ret;

	if (!realpath(prog, my_dir_name)) {
		res = 1;
		print_msg_warning("%s.%d Не удалось получить realpath от %s, "
			"res = NULL %s\n",
			FNC, LN, prog, strerror(errno));
	}
	if (vverbose)
		print_msg("prog =%s,"
			"\n\tПолный путь к программе =%s (%zu байт)\n",
			prog, my_dir_name, strlen(my_dir_name));

	for (n =0, i = strlen(my_dir_name); i > 0; i--) {
		if (my_dir_name[i] == '/') {
			if (++n == 1)
				my_prog_name = &(my_dir_name[i +1]);
			my_dir_name[i] = 0;
			break;
		}
	}

	if (verbose)
		print_msg("Дир. теста =%s, имя прог. =%s\n",
			my_dir_name, my_prog_name);
	ret = chdir(my_dir_name);
	if (ret) {
		print_msg_warning("%s.%d chdir(%s), рез. =%d %s\n",
			FNC, LN, my_dir_name, ret, strerror(errno));
		res = 1;
	}

	return (res);
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ set_current_dir */

extern int optind, opterr, optopt;
static int noarg, dbgopt = 0;

/*
 *   Пока не удалось при использованиии функции getopt получить информацию
 * об отсутствии аргумента в текущей опции,
 * хотя на сногочисленных сайтах описано как это "сделать"...
 *   функция checkopt позволяет "это сделать"...
 * Проверка наличия аргумента в текущей опции :
 *   результат функции =1, когда нет оператора.
 */
static int
checkopt () {
    if (!optarg)
        return 1;
    if (dbgopt)
	printf("%s :optarg =%s, fc =%c\n", FNC, optarg, *optarg);
    if (*optarg == '-') {
        optind--;
        noarg = 1;
        return 1;
    }
    return 0;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ checkopt */

static int
parse_parametrs (int argc, char **argv) {
	int	opt, res = 0, i;

    if ((argc > 0) && !strcmp(argv[1], "-d")) {
	dbgopt = 1;
	print_msg_note("По опции -d, включен режим =отладка getopt\n");
    }
    if (dbgopt)
	for (i = 1; i < argc; i++)
	    printf("\tN =%d\targ =%s\n", i, argv[i]);

    while ((opt = getopt(argc, argv, "hHe:C:dt:N:l:L:p:P:sS:wv")) != EOF) {
	if (dbgopt)
	    printf("opt =%c, optarg =%s, optind =%d, optopt =%c(%d), opterr =%d\n",
    		opt, optarg, optind, optopt, optopt, opterr);

	noarg = 0;
	switch (opt) {
	case	'd' : break;
	case	'p' :	// имя "подыгрывающей" программы
            if (!checkopt()) {
		if (prog_send)
		    prog_send2 = prog_send;
		else
		    prog_send = optarg;
		print_msg("Задано -имя \"подыгрывающей\" программы :%s\n",
		    prog_send);
	    }
	    break;
	case	'P' :	// N пина для "подыгрывающей" программы
            if (!checkopt()) {
		if (NpinP >= 0)
		    NpinP2 = atoi(optarg);
		else
		    NpinP = atoi(optarg);
		print_msg("Задано - N пина для \"подыгрывающей\" программы =%s\n",
		    optarg);
	    }
	    break;

	case	'e' :	// EXCL - задание на пины исключенные из тестирования
            if (!checkopt()) {
		str_excl = optarg;
		print_msg("Задано -исключить из тестирования пины ="
		    "\n\t\t'%s'\n", str_excl);
	    }
	    break;

	case	'w' :
    /*
	режим ожидания прерываний :
	    wait_regime = 1 - переднему фронту "rising";
	    wait_regime = 2 - заднему   фронту "falling";
	    wait_regime > 2 - по обеим фронтам "both"
     */
	    wait_regime++;
	    break;

	case	't' :	// время работы теста (в секундах)
            if (!checkopt()) {
		wait_time = atol(optarg);
		print_msg("Задано -время тестирования =%d сек.\n", wait_time);
	    }
	    break;

	case	'C' : // предел числа ожидаемых прерываний (для отладки)
            if (!checkopt()) {
		number_checks = atol(optarg);
		print_msg("Задано -число ожидаемых прерываний =%d\n",
		    number_checks);
            }
	    break;

	case	'N' : // ограничение оператором числа пинов gpio
            if (!checkopt())
                NGPIO = atol(optarg);
            break;

	case	'l' :
            if (!checkopt())
                limit_error = atol(optarg);
            break;

	case	'L' :	// число ошибок errpoll
            if (!checkopt())
                limit_errpoll = atol(optarg);
            break;

	case	's' :	// без контроля запаяных пинов (для отладки)
	    without_check_solder = 1;
	    print_msg("Задано -тестировать без контроля "
		"'запаяных пинов' !\n");
	    break;

    /*
     * (sleep, для отладки)
     * задержка в милисек. перед ожиданием прерываний
     */
	case	'S' :
            if (!checkopt()) {
		timeout_before_check = atol(optarg);
		print_msg_note("Задано -перед ожиданием прерываний произвести "
		    "задержку в %d милисек.\n", timeout_before_check);
	    }
            break;

	case	'v' :	// v|vv|vvv - отладочная печать
	    verbose++;
	    break;

	case	'?' :
	    print_msg_error(" ? в задании параметров !\n");
	    usage_and_exit = 2;
	    break;
	case	':' :
	    print_msg_error(" : в задании параметров !\n");
	    usage_and_exit = 2;
	    break;
	case	'h' :
	    usage_and_exit = 0;
	    break;
	case	'H' :
	    usage_and_exit = 1;
	    break;
	default :
	    usage_and_exit = 2;
	    break;
	}

    if (noarg) {
	print_msg_error(" !!! опция -%c - нет аргументов !\n", opt);
	res++;
    }

    if (optarg)
	sprintf(&cmdline[strlen(cmdline)], "-%c %s ", (char)opt, optarg);
    else
	sprintf(&cmdline[strlen(cmdline)], "-%c ", (char)opt);
    }
    if (res)
	usage_and_exit = 2;
    return res;
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ parse_parametrs */

int
main (int argc, char *argv[]) {
	int	i, j;

	set_current_dir(argv[0]);
	sprintf(&cmdline[strlen(cmdline)], "%s ", argv[0]);
	t_start_tst =	// на случай если тест и "не запустится"...
	t_start_prog = Gethrtime();
	my_pid = getpid();

	parse_parametrs(argc, argv);
	if (usage_and_exit >= 0) {
	    usage(usage_and_exit);
	    exit(usage_and_exit);
	}
	if (prog_send && access(prog_send, F_OK)) {
    	    print_msg_error("!!!нет файла =%s !\n", prog_send);
    	    exit(1);
	}
	if (prog_send2 && access(prog_send, F_OK)) {
    	    print_msg_error("!!!нет файла N 2 =%s !\n", prog_send2);
    	    exit(1);
	}

	ngpio = define_ngpio();
	if (NGPIO > ngpio) {
	    print_msg_warning(" NGPIO =%d > ngpio =%d, set NGPIO =%d\n",
		NGPIO, ngpio, ngpio);
	    NGPIO = ngpio;
	}
	if (NGPIO < 0)
	    NGPIO = ngpio;

	if (str_excl) {
	    if (parse_str_excl(str_excl) || (nrange < 0)) {
        	print_msg_error("\t!!!\n"
		    "\tПри разборе задания исключений ='%s' !\n", str_excl);
        	exit(1);
    	    }
	    if (parse_excl_pins())
		exit(1);
	}
        if (NpinP >=0)
            if (!PIN_NEVER_EXPORT(NpinP)) {
                    print_msg_note(":пин N %2d исключаем !\n", NpinP);
                PIN_NEVER(NpinP) = PIN_NEVER_EXPORT(NpinP) = 1;
            }
        if (NpinP2 >=0)
            if (!PIN_NEVER_EXPORT(NpinP2)) {
                    print_msg_note(":пин N %2d исключаем !\n", NpinP2);
                PIN_NEVER(NpinP2) = PIN_NEVER_EXPORT(NpinP2) = 1;
            }

	start_t(NM_TST);
	set_sig_handler(SIGINT, sigcommon_hndlr);
	set_sig_handler(SIGALRM, sigcommon_hndlr);

	/*
	 *    Печать строки запуска, если verbose
	 * или запуск не из (СТДП) tst_run, ...
	 */
	{
	char *T = getenv("MY_NUMBER");
		if (T == NULL || verbose)
			print_msg("RUN :%s\n", cmdline);
	}

	if (verbose)
	    print_msg("define_ngpio :ngpio из %s/gpiochip*/ngpio=%d\n",
		DGPIO, ngpio);

	print_msg_note("DGPIO=%s\n", DGPIO);

	/*
	 *	Убираем ранее (кем-то) export-ированные gpio
	 */
	for (i = 0; i < NGPIO; i++) {
		char nmgpio[MAX_SYSFS_NAME_LEN];
		sprintf(nmgpio, "%s/%s%d", DGPIO, pins_name, i);
		if (!access(nmgpio , F_OK)) { /* есть файл gpio<пин> */
			if (vvverbose)
				print_msg_note("Есть %s\n", nmgpio);
			gpio_pin_close(i);
		}
	}

	/*
	 *    Открытие пино-ов GPIO и инициализация уровня PIN_ACTIVE_LOW,
	 * установка PIN_NEVER и PIN_AVAIL.
	 */
	{
		char msg[512], *msg_ = msg;

		for (j = i = 0; i < NGPIO; i++) {
			if (!pin_is_valid(i)) {
				PIN_NEVER(i) = 1;
				continue;
			}

			PIN_AVAIL(i) = gpio_pin_open(i) ? 0 : 1;
			if (!PIN_AVAIL(i)) {
				print_msg_error("Не удалось открыть SYSFS "
					"для пина %d\n", i);
				errors++;
				continue;
			}

			PIN_ACTIVE_LOW(i) = gpio_pin_get_active_low(i);
			if (!PIN_NEVER(i)) {
				msg_ += sprintf(msg_, "%d =%d, ",
						i, PIN_ACTIVE_LOW(i));
				j++;
			}
		}
		if (j) {
			msg_[-2] = 0;
			if (verbose)
				print_msg("Значения active_low:\n\t%s\n", msg);
		}
		if (vverbose) {
			char cmd[128];
			sprintf(cmd, "find %s | grep %s/ | sed 's/^/\t/'",
				DGPIO, DGPIO);
			print_msg("Дерево sysfs-GPIO :\n");
			system(cmd); 
		}
	}

	check_solder();

	if (init_poll())
	    goto out_tst;

	if (verbose) {
	    char msg[1024], *msg_ = msg;
	/*
	 *  Печать пинов исключенных из тестирования
	 */
	    for (j = 0, i = 0; i < NGPIO; i++) {
		if (PIN_NEVER_EXPORT(i)) {
		    msg_ += sprintf(msg_, "%d, ", i);
		    j++;
		}
	    }
	    if (j) {
		msg_[-2] = 0;
		if (fname == NULL)
		    print_msg_note("\n\tПины запрещенные для тестирования :"
			"%s%s\n", j > 4 ? "\n\t\t" : "", msg);
		else
		    print_msg_note("\n\tПины запрещенные для тестирования "
			"(см.парам-файл =%s) :%s%s\n", fname,
			j > 4 ? "\n\t\t" : "", msg);
	    }

	    prn_state();
	}
	t_start_tst = Gethrtime();
	wait_inters();

out_tst:
	close_exit();
}
