/*
    wait_one_gpio.c - gpio-тест ожидания прерываний на заданных линиях
    разработчик - Т.Гумер

    автономная сборка :
gcc -g -o wait_one_gpio wait_one_gpio.c

    инструкция использования :запуск программы без параметров.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <signal.h>

#define DGPIO "/sys/class/gpio"

static int	Npin = -1;
static struct	pollfd pollfd[1];	// pollfd для операций poll

/*
 * Все проверяемые пины настроим на прием прерывания
 * по :
    wait_regime = 1 - переднему фронту "rising";
    wait_regime = 2 - заднему   фронту "falling";
    wait_regime > 2 - по обеим фронтам "both"
 */
static int wait_regime = 1;	// rising

/*
    #define POLLIN      0x0001    Можно считывать данные
    #define POLLPRI     0x0002    Есть срочные данные
    #define POLLOUT     0x0004    Запись не будет блокирована
    #define POLLERR     0x0008    Произошла ошибка
    #define POLLHUP     0x0010   "Положили трубку"
    #define POLLNVAL    0x0020    Неверный запрос: fd не открыт
 */

//#define HRT_SEC(SEC)    ((hrtime_t)(SEC *1000000000LL))

static struct timeval start_time;
static struct timeval cur_time;
static float
time_mlsec () {
    static int inn = 0;

    if (!inn++) {
	gettimeofday(&start_time, 0);
    }
    gettimeofday(&cur_time, 0);
    return (float)(cur_time.tv_sec*1000000LL +cur_time.tv_usec
		-start_time.tv_sec*1000000LL -start_time.tv_usec)/1000000LL;
}

static float test_run_time, t_start_prog, t_start_tst;
static int fd = -1, Nintr = 0, Nerr_poll = 0;
static int fexport = -1;

static void
close_exit () {
    char cmd[512];
    if (fexport > 0)
	close(fexport);
    if (fd > 0)
	close(fd);

    sprintf(cmd, "echo %d > %s/unexport", Npin, DGPIO);
//	system("set +x; find /sys/class/gpio");
    if (1)
	printf("run :%s\n", cmd);
    system(cmd);
//	system("set +x; find /sys/class/gpio");

    printf("Время работы :теста =%5.3f, программы =%5.3f сек.\n"
	"Получено %d прерываний (N пина =%d).\n",
	test_run_time, time_mlsec(), Nintr, Npin);
    exit (0);
} /* END ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ close_exit */

static void
ctrlc (int sig) {
    printf("%s :Нажато '<Control+C>' сигнал N %d\n", __FUNCTION__, sig);
    close_exit();
}

int
main (int argc, char *argv[]) {
    char cmd[128], value_name[128];
    int	i, j, start_play_along_program = 0, NPIN = -1;
    u_long all_checks;
    FILE *fvalue;
    t_start_prog = time_mlsec();

    if (argc <= 1) {
	printf("Используй :%s <Npin> [<режим ожидания =1|2|3> [<NPIN>] ]"
	    "\n   где :"
	    "\n\tNpin - N линии ожидания сигналов gpio;"
	    "\n\tрежим ожидания	= 1 - ждать по переднему фронту (по умолчанию)"
	    "\n\tрежим ожидания	= 2 - ждать по заднему   фронту"
	    "\n\tрежим ожидания	> 2 - по обеим фронтам;"
	    "\n\tNPIN - запуск 'подыгрывающего скрипта' = ./CMDone NPIN\n\n",
	    argv[0]);
	exit (1);
    }

    Npin = atoi(argv[1]);
    printf("Старт %s :ожидание сигналов gpio на линии gpio N %d\n", argv[0], Npin);
    if (argc > 2) {
	wait_regime = atoi(argv[2]);
	if (wait_regime == 1)
	    printf("\tрежим ожидания =ждать по переднему фронту\n");
	if (wait_regime == 2)
	    printf("\tрежим ожидания =ждать по заднему фронту\n");
	if (wait_regime >= 3)
	    printf("\tрежим ожидания =ждать по обеим фронтам\n");
    }
    if (argc > 3) {
	printf("set regime 'start play-along program'\n");
	start_play_along_program = 1;
	NPIN = atoi(argv[3]);
    }

    { // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_open
	char *F = "gpio_pin_open";
	char export_name[1024];
	char gpio_name[1024];
	int szNpin, rwrite;
	char mNpin[64];

        sprintf(export_name, "%s/export", DGPIO);
        fexport = open(export_name, O_WRONLY);

        if (fexport < 0) {
    	    printf("%s :!!! Ошибка открытия %s - %m\n", F, export_name);
	    exit(1);
        }
	printf("%s :Файл = %s - открыт\n", F, export_name);

	sprintf(gpio_name, "%s/gpio%d", DGPIO, Npin);
	if (access(gpio_name , F_OK)) { /* нет файла gpio<пин> */
	    sprintf(mNpin, "%d", Npin);
	    rwrite = write(fexport, mNpin, szNpin = strlen(mNpin));
	    if (rwrite != szNpin) {
    		printf("%s :Ошибка export-ирования %s - %m\n", F, gpio_name);
		exit (0);
	    }
	/*
	 * Проверка export-ирования линии GPIO
	 */
	    if (!access(gpio_name , F_OK)) { /* появился файл gpio<пин> */
        	printf("%s :линия GPIO =%s\texport-ирована\n", F, gpio_name);
    	    } else {
    		printf("%s :Нет линии = %s\n", F, gpio_name);
		exit(1);
	    }
	} else
            printf("%s :линия GPIO =%s\tУЖЕ export-ирована\n", F, gpio_name);

    } // end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_open

    { // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ set direction
	char min[64];
	sprintf(min, "echo in > %s/gpio%d/direction", DGPIO, Npin);
	system(min);
	printf("Установить текущее состояние линии %s/gpio%d/direction ='in'\n",
	    DGPIO, Npin);
    } // end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ set direction

    sprintf(value_name, "%s/gpio%d/value", DGPIO, Npin);
    fd = open(value_name, O_RDWR | O_NONBLOCK);
    if (fd == -1) {
	printf("Err open %s - %m\n", value_name);
	exit (1);
    }
    if (wait_regime > 2) {
        fvalue = fopen(value_name, "r");
        if (!fvalue) {
    	    printf("!!! ОШИБКА !!! открытия %s\n", value_name);
	    exit (1);
        }
    }

    if (start_play_along_program) {
	char cmd[128];
	sprintf(cmd, "export PATH=.:$PATH; ./CMDone %lld %d &",
	    (long long)(start_time.tv_sec*1000LL +
			start_time.tv_usec/1000000LL), NPIN);
	system(cmd);
    }

    { // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_set_edge
	char name_edge[1024], *sedge;
        FILE *file_edge;
        sprintf(name_edge, "%s/gpio%d/edge", DGPIO, Npin);
        file_edge = fopen(name_edge, "w");
        if (!file_edge) {
                printf("!!! Ошибка открытия %s - %m\n", name_edge);
                exit (1);
        }
	sedge = wait_regime == 1 ? "rising" :
                wait_regime == 2 ? "falling" : "both";
        fputs(sedge, file_edge);
	printf("gpio_pin_set_edge :'%s' -> %s\n", sedge, name_edge);
        fclose(file_edge);
    } // end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_set_edge

    pollfd[0].fd = fd;
    t_start_tst = time_mlsec();
    signal(SIGINT, ctrlc);	// для перехвата 
    for (all_checks = 0; Nerr_poll < 3; all_checks++) {
        int Nrevents, revents, time_out = 10000, value, poll_res;
	char *mval = "", buf[16] = {'y'};;

if (Nintr > 10)	// ZZZ
    break;
	pollfd[0].events = POLLPRI;
	pollfd[0].revents = 0;
        read(fd, buf, sizeof(buf));	// reset_irqs
        poll_res = poll(pollfd, 1, time_out);  // ож. 10 сек.
        test_run_time = time_mlsec();

        if (!poll_res) {
            printf("%5.3f :\tpoll :time_out =%d mlsec ожидания прер.\n",
                    test_run_time, time_out);
	    continue;
        }
        if (poll_res < 0) {
            printf("%5.3f :\t!!! ОШИБКА ожидания прер. - %m\n", test_run_time);
            Nerr_poll++;
	    continue;
        }
	Nintr++;
	if (wait_regime == 3) { // ~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_get_value
    	    char value_name[1024];
    	    char str[16];
    	    FILE *value = NULL;
	    int val;

    	    sprintf(value_name, "%s/gpio%d/value", DGPIO, Npin);
    	    value = fopen(value_name, "r");
    	    if (!value)
    		printf("gpio_pin_get_value :!!! ОШИБКА !!! открытия %s -%m\n",
		    value_name);

    	    fgets(str, 2, value);
    	    fclose(value);

    	    if (0)
                printf("val %s\t@%s\n", value_name, str);
    	    val = atoi(str);
	    mval = val ? "+" : "-";
	} // end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ gpio_pin_get_value
	printf("%5.3f :\tесть прер. N %d %s\n", test_run_time, Nintr, mval);
    }

close_exit();
}
